"""
pangaeapy is a package allowing to download and analyse metadata
as well as data from tabular PANGAEA (https://www.pangaea.de) datasets.
"""
from pangaeapy.pandataset import PanAuthor,PanEvent,PanParam,PanDataSet
from pangaeapy.panquery import PanQuery